import {required} from "./../../../../node_modules/pragma-core/lib/utils.js";
import * as StorageConstants from "./storage-constants.js";
import {IndexedDbResponse} from "./indexedDb-response.js";
import {LocalStoreBase} from "./local-store-base.js";

export class LocalStore extends LocalStoreBase {    
    
    constructor(databaseName, databaseVersion, migrator) {
        super();
        
        required(databaseName);        
        required(databaseVersion);        
        required(migrator);
        
        this._name = databaseName;
        this._version = databaseVersion;
        this._migrator = migrator;       
    }

    async _createConnection(callbackAsync) {
        let response = new IndexedDbResponse();
        let connection = null;
        try {
            const connectionResponse = await this._openConnection();
            if (connectionResponse.isValid()) {
                connection = connectionResponse.result;
                response = await callbackAsync(connection);
            }
        }
        catch (exception) {                
            response.notification.addError(StorageConstants.Errors.IndexedDbGeneral, `Unhandled IndexedDb exception:\r\n ${exception.stack}`);
        }
        finally {
            if (connection != null) {
                connection.close();
                connection = null;
            }
        }
        
        return response;
    }

    async _onUpgradeNeededHandler(event, response, promiseResolve) {
        const upgradeTransaction =  event.target.transaction;
        upgradeTransaction.oncomplete = () => promiseResolve(response);
        
        const migrationResponse = await this._migrator.migrateAsync(event.target.result, event.oldVersion);
        if (migrationResponse.isValid()) {
            response.result = event.target.result;
        }
    }
    
    async _openConnection() {
        return new Promise(resolve =>  {
            const response = new IndexedDbResponse();
            const openRequest = indexedDB.open(this._name, this._version);
            openRequest.onerror = event => this._onErrorHandler(event, response, resolve);
            openRequest.onsuccess = event => this._onSuccessHandler(event, response, resolve);
            openRequest.onupgradeneeded = event => this._onUpgradeNeededHandler(event, response, resolve);            
        });
    }
    
    async clearStore(storeName) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName, "readwrite");
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);
                transaction.oncomplete = event => this._onSuccessHandler(event, response, resolve);
                transaction.onabort = event => this._onErrorHandler(event, response, resolve);
                
                const store = transaction.objectStore(storeName);                
                store.clear();
            });
        });
    }

    async count(storeName, dbKeyRange) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName);
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);
                
                const store = transaction.objectStore(storeName);
                const request = store.count(dbKeyRange);
                request.onsuccess = event => this._onSuccessHandler(event, response, resolve);
            })
        });        
    }

    async delete(storeName, keys) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {     
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName, "readwrite");
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);
                transaction.oncomplete = () => resolve(response);
                transaction.onabort = event => this._onErrorHandler(event, response, resolve);
    
                const store = transaction.objectStore(storeName);
                for (let key of keys) {
                    store.delete(key);
                }
            });            
        });
    }

    async deleteDatabase() {
        return new Promise(resolve => {
            const response = new IndexedDbResponse();
            const deleteRequest = indexedDB.deleteDatabase(this._name);
            deleteRequest.onerror = event => this._onErrorHandler(event, response, resolve);
            deleteRequest.onsuccess = event => this._onSuccessHandler(event, response, resolve);
            deleteRequest.onblocked = event => this._onBlockedHandler(event, response, resolve);
        });
    }
    
    async query(storeName, filterCallback, howMany) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName);
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);

                let count = 0;
                // noinspection JSMismatchedCollectionQueryUpdate
                const queriedRecords = [];
                const store = transaction.objectStore(storeName);
                const openCursor = store.openCursor();
                openCursor.onerror = event => this._onErrorHandler(event, response, resolve);
                openCursor.onsuccess = event => {
                    const cursor = event.target.result;
                    if (cursor != null) {
                        const record = cursor.value;
                        if (filterCallback != null) {
                            if (filterCallback(record)) {
                                queriedRecords.push(record);
                                count++;
                            }
                        }
                        else {
                            queriedRecords.push(record);
                            count++;
                        }

                        if ((howMany == null) || (count < howMany)) {
                            cursor.continue();
                        }
                        else {
                            response.result = queriedRecords;
                            resolve(response);                            
                        }
                    }
                    else {
                        response.result = queriedRecords;
                        resolve(response);
                    }
                };
            });
        });
    }

    async queryAll(storeName, dbKeyRange, count) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName);
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);
                
                const store = transaction.objectStore(storeName);
                const request = store.getAll(dbKeyRange, count);
                request.onsuccess = event => this._onSuccessHandler(event, response, resolve);
            })
        });        
    }
    
    async queryAllKeys(storeName, dbKeyRange, count) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName);
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);
                
                const store = transaction.objectStore(storeName);
                const request = store.getAllKeys(dbKeyRange, count);
                request.onsuccess = event => this._onSuccessHandler(event, response, resolve);
            })
        });        
    }

    async save(storeName, records) {
        return this._createConnection(async connection => {
            return new Promise(resolve => {     
                const response = new IndexedDbResponse();
                const transaction = connection.transaction(storeName, "readwrite");
                transaction.onerror = event => this._onErrorHandler(event, response, resolve);
                transaction.oncomplete = () => resolve(response);
                transaction.onabort = event => this._onErrorHandler(event, response, resolve);
    
                const store = transaction.objectStore(storeName);
                for (let record of records) {
                    store.put(record);
                }
            });            
        });
    }   


}